@@ -1,9 +1,12 @@
 import inspect
 import types
+from datetime import date
+from enum import Enum, IntEnum
+from pathlib import Path
 from typing import Any
 
 
-def dict_has_key_chain(source: dict[Any, Any],
+def dict_has_key_chain(source: dict,
                        key_chain: list[str]) -> bool:
     """
     Indicate the existence of an element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.
@@ -54,8 +57,8 @@ def dict_has_key_chain(source: dict[Any, Any],
     return result
 
 
-def dict_get_value(source: dict[Any, Any],
-                   key_chain: list[str]) -> any:
+def dict_get_value(source: dict,
+                   key_chain: list[str]) -> Any:
     """
     Obtain the value of the element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.
 
@@ -71,7 +74,7 @@ def dict_get_value(source: dict[Any, Any],
     :return: the value obtained
     """
     # initialize the return variable
-    result: any = source
+    result: Any = source
 
     # traverse the keys in the chain
     for key in key_chain:
@@ -107,7 +110,7 @@ def dict_get_value(source: dict[Any, Any],
 
 def dict_set_value(target: dict,
                    key_chain: list[str],
-                   value: Any) -> None:
+                   value: Any) -> dict:
     """
     Assign to an element of *source* the value *value*.
 
@@ -116,12 +119,14 @@ def dict_set_value(target: dict,
     Any non-existing intermediate elements are created with the value of an empty *dict*.
     A key might indicate the position of the element within a list, using the format *<key>[<pos>]*.
     In such a case, that element must exist.
+    For convenience, the possibly modified *target* itself is returned.
 
     :param target: the reference dict
     :param key_chain: the key chain
     :param value: the value to be assigned
+    :return: the modified input *dict*
     """
-    dict_item: any = target
+    dict_item: Any = target
     # traverse the chain, up to end including its penultimate element
     for key in key_chain[:-1]:
 
@@ -168,9 +173,11 @@ def dict_set_value(target: dict,
             # no, assign 'value' to the element 'key' in the dictionary
             dict_item[key] = value
 
+    return target
 
-def dict_pop_value(target: dict[Any, Any],
-                   key_chain: list[str]) -> any:
+
+def dict_pop_value(target: dict,
+                   key_chain: list[str]) -> Any:
     """
     Obtain the value of the element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.
 
@@ -186,7 +193,7 @@ def dict_pop_value(target: dict[Any, Any],
     :return: the value removed
     """
     # initialize the return variable
-    result: any = None
+    result: Any = None
 
     # obtain the element pointed to by the las key in the chain
     parent: dict | None = None
@@ -213,7 +220,7 @@ def dict_pop_value(target: dict[Any, Any],
             pos: int = key.find("[")
             inx: int = int(key[pos+1:-1])
             key = key[:pos]
-            child: any = parent.get(key)
+            child: Any = parent.get(key)
 
             # is the element pointed to by the last key in the chain a list with more than 'inx' elements ?
             if isinstance(child, list) and len(child) > inx:
@@ -228,36 +235,36 @@ def dict_pop_value(target: dict[Any, Any],
     return result
 
 
-def dict_replace_value(target: dict[Any, Any],
-                       old_value: any,
-                       new_value: any) -> None:
+def dict_replace_value(target: dict,
+                       old_value: Any,
+                       new_value: Any) -> dict:
     """
     Replace, in *target*, all occurrences of *old_value* with *new_value*.
 
+    For convenience, the possibly modified *target* itself is returned.
+
     :param target: the reference dict
     :param old_value: the value to be replaced
     :param new_value: the new value
+    :return: the modified input *dict*
     """
-    def list_replace_value(items: list[any],
-                           old_val: any,
-                           new_val: any) -> None:
+    def list_replace_value(items: list,
+                           old_val: Any,
+                           new_val: Any) -> None:
         # traverse the list
         for item in items:
-
             # is 'item' a dict ?
             if isinstance(item, dict):
                 # yes, process it recursively
                 dict_replace_value(target=item,
                                    old_value=old_val,
                                    new_value=new_val)
-
             # is 'item' a list ?
             elif isinstance(item, list):
                 # yes, process it recursively
                 list_replace_value(items=item,
                                    old_val=old_val,
                                    new_val=new_val)
-
     # traverse the dict
     for curr_key, curr_value in target.items():
 
@@ -279,21 +286,36 @@ def dict_replace_value(target: dict[Any, Any],
             list_replace_value(items=curr_value,
                                old_val=old_value,
                                new_val=new_value)
+    return target
 
 
-def dict_get_key(source: dict[Any, Any],
-                 value: any) -> any:
+def dict_unique_values(source: dict) -> list:
+    """
+    Return a list with the values in *source*, pruned of duplicates.
+
+    The elements in the returned list maintain the original insertion order in *source*.
+
+    :param source: the dict to retrieve the values from
+    :return: list the values in the source *dict*, pruned of duplicated
+    """
+    # a 'dict' maintains the insertion order of its elements
+    uniques: dict[Any, None] = dict.fromkeys(source.values())
+    return list(uniques.keys())
+
+
+def dict_get_key(source: dict,
+                 value: Any) -> Any:
     """
     Return the key in *source*, mapping the first occurrence of *value* found.
 
-    Return *None*, if no key is found.
     No recursion is attempted; only the first-level attributes in *source* are inspected.
+    Return *None*, if no key is found.
 
     :param source: dict to search
     :param value: the reference value
     :return: first key mapping the reference value
     """
-    result: any = None
+    result: Any = None
     for key, val in source.items():
         if val == value:
             result = key
@@ -302,35 +324,46 @@ def dict_get_key(source: dict[Any, Any],
     return result
 
 
-def dict_get_keys(source: [Any,Any],
-                  value: any) -> list[str]:
+def dict_get_keys(source: dict,
+                  value: Any) -> list:
     """
     Return all keys in *source*, mapping the value *value*.
 
+    The search is done recursively. Note that *dict*s in *list*s are not searched.
+    The order of the keys returned should not be taken as relevant.
     Return *[]* if no key is found.
 
     :param source: dict to search
     :param value: the reference value
     :return: list containing all keys mapping the reference value
     """
-    return [key for key, val in source.items() if val == value]
+    # initialize the return variable
+    result: list = []
+    for item_key, item_value in source.items():
+        if item_value == value:
+            result.append(item_key)
+        elif isinstance(item_value, dict):
+            result.extend(dict_get_keys(source=item_value,
+                                        value=value))
+    return result
 
 
-def dict_merge(target: dict[Any, Any],
-               source: dict) -> None:
+def dict_merge(target: dict,
+               source: dict) -> dict:
     """
     Traverse the elements in *source* to update *target*, according to the criteria presented herein.
 
     The criteria to be followed are:
       - add the element to *target*, if it does not exist
       - if the element exists in *target*:
-
         - recursively process both elements, if both are type *dict*
         - add the missing items, if both are type *list*
         - replace the element in *target* if it is a different type, ou if both elements are not of the same type
+    For convenience, the possibly modified *target* itself is returned.
 
     :param target: the dictionary to be updated
     :param source: the dictionary with the new elements
+    :return: the modified target *dict*
     """
     # traverse the dictionary with the new elements
     for skey, svalue in source.items():
@@ -338,7 +371,7 @@ def dict_merge(target: dict[Any, Any],
         # is the item in target ?
         if skey in target:
             # yes, proceed
-            tvalue: any = target.get(skey)
+            tvalue: Any = target.get(skey)
 
             # are both elements dictionaries  ?
             if isinstance(svalue, dict) and isinstance(tvalue, dict):
@@ -359,17 +392,21 @@ def dict_merge(target: dict[Any, Any],
             # the item is not in target, add it
             target[skey] = svalue
 
+    return target
 
-def dict_coalesce(target: dict[Any, Any],
-                  key_chain: list[str]) -> None:
+
+def dict_coalesce(target: dict,
+                  key_chain: list[str]) -> dict:
     """
     Coalesce the element of type *list* in *target* at the level *n* with the list at the level immediately above.
 
     This element is pointed to by the key chain *[keys[0]: ... :keys[n]*, and is processed as a sequence
     of multiple elements. The two last keys in *key_chain* must be associated with values of type *list*.
+    For convenience, the possibly modified *target* itself is returned.
 
     :param target: the dictionary to be coalesced
     :param key_chain: the chain of nested keys
+    :return: the modified input *dict*
     """
     # traverse the kay chain up to its penultimate element
     curr_dict: dict | None = target
@@ -382,7 +419,7 @@ def dict_coalesce(target: dict[Any, Any],
             break
 
         # is 'key' associated to a list ?
-        in_list: list[any] = curr_dict.get(key)
+        in_list: list = curr_dict.get(key)
         if isinstance(in_list, list):
             # yes, recursively invoke the coalescing of the dictionaries in the list
             for in_dict in in_list:
@@ -415,24 +452,22 @@ def dict_coalesce(target: dict[Any, Any],
                 last_list: list[dict] = []
 
                 # traverse the last element
-                for k1, v1 in last_elem.items():
-                    # if 'k1' the last key, and is it a list ?
-                    if k1 == key_chain[-1] and isinstance(v1, list):
+                for k, v in last_elem.items():
+                    # if 'k' the last key, and is it a list ?
+                    if k == key_chain[-1] and isinstance(v, list):
                         # yes, obtain its items for further coalescing
-                        for in_dict in v1:
+                        for in_dict in v:
                             # is 'in_dict' a dictionary ?
                             if isinstance(in_dict, dict):
                                 # yes, coalesce and save it
-                                inner_dict: dict = {}
-                                for k2, v2 in in_dict.items():
-                                    inner_dict[k2] = v2
+                                inner_dict: dict = dict(in_dict.items())
                                 last_list.append(inner_dict)
                             else:
                                 # no, save it as is
                                 last_list.append(in_dict)
                     else:
                         # no, coalesce it
-                        outer_dict[k1] = v1
+                        outer_dict[k] = v
 
                 # are there items to be coalesced ?
                 if len(last_list) > 0:
@@ -454,17 +489,21 @@ def dict_coalesce(target: dict[Any, Any],
         # replace the original list with the coalesced new list
         curr_dict[key_chain[-2]] = penultimate_list
 
+    return target
 
-def dict_reduce(target: dict[Any, Any],
-                key_chain: list[str]) -> None:
+
+def dict_reduce(target: dict,
+                key_chain: list[str]) -> dict:
     """
     Relocate the elements from *target* at level *n*, to the level immediately above.
 
     These elements are pointed to by the nested key chain *[keys[0]: ... :keys[n]*.
     The element at level *n* is removed at the end.
+    For convenience, the possibly modified *target* itself is returned.
 
     :param target: the 'dict' to be reduced
     :param key_chain: the key chain
+    :return: the modified input *dict*
     """
     # does the key chain contain at least 1 key ?
     if len(key_chain) > 0:
@@ -480,7 +519,7 @@ def dict_reduce(target: dict[Any, Any],
                 break
 
             # is 'key' associated with a list ?
-            in_list: list[any] = curr_dict.get(key)
+            in_list: list = curr_dict.get(key)
             if isinstance(in_list, list):
                 # yes, recursively invoke reduction of the dictionaries in 'in_list'
                 for in_dict in in_list:
@@ -505,17 +544,19 @@ def dict_reduce(target: dict[Any, Any],
             for key, value in last.items():
                 curr_dict[key] = value
 
+    return target
 
-def dict_from_list(source: list[dict[Any, Any]],
+
+def dict_from_list(source: list[dict],
                    key_chain: list[str],
-                   value: any) -> dict:
+                   value: Any) -> dict:
     """
     Locate in *source*, and return, the element of type *dict* with value *value* in the key chain *key_chain*.
 
     :param source: the list to be inspected
     :param key_chain: the key chain used in the search process
     :param value: the value of the element pointed to by the key chain
-    :return: the 'dict' wanted, or 'None' if not found
+    :return: the *dict* wanted, or *None* if not found
     """
     # initialize the return variable
     result: dict | None = None
@@ -537,7 +578,7 @@ def dict_from_object(source: object) -> dict:
     The input *source* might be any *object*, specially those decorated with *@dataclass*.
 
     :param source: the reference object
-    :return: 'dict' structurally similar to the reference object
+    :return: *dict* structurally similar to the reference object
     """
     # initialize the return variable
     result: dict = {}
@@ -577,8 +618,8 @@ def dict_from_object(source: object) -> dict:
     return result
 
 
-def dict_transform(source: dict[Any, Any],
-                   from_to_keys: list[tuple[str,str]],
+def dict_transform(source: dict,
+                   from_to_keys: list[tuple[str, Any]],
                    prefix_from: str = None,
                    prefix_to: str = None) -> dict:
     """
@@ -588,8 +629,7 @@ def dict_transform(source: dict[Any, Any],
     *from_to_keys*, the element indicated by the second term of the tuple, assigning to it
     the value of the *source* element indicated by the first term of the tuple. Both terms
     of the tuples are represented by a chain of nested keys.
-
-    The prefixes for the source and destination keys, if defined, have fferent treatments.
+    The prefixes for the source and destination keys, if defined, have different treatments.
     They are added when searching for values in *Source*, and removed when assigning values
     to the return *dict*.
 
@@ -597,10 +637,10 @@ def dict_transform(source: dict[Any, Any],
     :param from_to_keys: the list of tuples containing the source and destination key sequences
     :param prefix_from: prefix to be added to source keys
     :param prefix_to: prefix to be removed from target keys
-    :return: the new 'dict'
+    :return: the new *dict*
     """
     # import the needed functions
-    from .list_pomes import list_find_coupled, list_transform, list_unflatten
+    from .list_pomes import list_get_coupled, list_transform, list_unflatten
 
     # initialize the return variable
     result: dict = {}
@@ -615,8 +655,8 @@ def dict_transform(source: dict[Any, Any],
             from_keys: str = key
 
         # get the target key chain
-        to_keys: str = list_find_coupled(coupled_elements=from_to_keys,
-                                         primary_element=from_keys)
+        to_keys: str = list_get_coupled(coupled_elements=from_to_keys,
+                                        primary_element=from_keys)
 
         # has the destination been defined ?
         if to_keys:
@@ -635,7 +675,7 @@ def dict_transform(source: dict[Any, Any],
                                                 prefix_to=to_keys)
             else:
                 # 'value' is neither a dictionary nor a list
-                to_value: any = value
+                to_value: Any = value
 
             # has the target prefix been defined and does it occur in the target string ?
             if prefix_to and to_keys.startswith(prefix_to):
@@ -652,7 +692,7 @@ def dict_transform(source: dict[Any, Any],
 
 
 def dict_clone(source: dict,
-               from_to_keys: list[str | tuple[str,str]]) -> dict:
+               from_to_keys: list) -> dict:
     """
     Build a new *dict*, according to the rules presented herein.
 
@@ -666,7 +706,7 @@ def dict_clone(source: dict,
 
     :param source: the source dict
     :param from_to_keys: list of elements indicative of the source and target keys
-    :return: the new dict
+    :return: the new *dict*
     """
     # import the needed function
     from .list_pomes import list_unflatten
@@ -680,24 +720,24 @@ def dict_clone(source: dict,
         to_key: str = (elem[1] if isinstance(elem, tuple) and len(elem) > 1 else None) or from_key
         result[to_key] = dict_get_value(source=source,
                                         key_chain=list_unflatten(from_key))
-
     return result
 
 
-def dict_listify(target: dict[Any, Any],
-                 key_chain: list[str]) -> None:
+def dict_listify(target: dict,
+                 key_chain: list[str]) -> dict:
     """
     Insert the value of the item pointed to by the key chain *[keys[0]: ... :keys[n]* in a list.
 
     This insertion will happen only if such a value is not itself a list.
     All lists eventually found, up to the penultimate key in the chain, will be processed recursively.
+    For convenience, the possibly modified *target* itself is returned.
 
     :param target: the dictionary to be modified
     :param key_chain: the chain of nested keys pointing to the item in question
+    :return: the modified input *dict*
     """
     def items_listify(in_targets: list,
                       in_keys: list[str]) -> None:
-
         # traverse the list
         for in_target in in_targets:
             # is the element a dictionary ?
@@ -712,7 +752,7 @@ def dict_listify(target: dict[Any, Any],
                 items_listify(in_targets=in_target,
                               in_keys=in_keys)
 
-    parent: any = target
+    parent: Any = target
     # traverse the chain up to its penultimate key
     for inx, key in enumerate(key_chain[:-1]):
         parent = parent.get(key)
@@ -733,9 +773,156 @@ def dict_listify(target: dict[Any, Any],
         # does the item exist and is not a list ?
         if key in parent and not isinstance(parent.get(key), list):
             # yes, insert it in a list
-            item: any = parent.pop(key)
+            item: Any = parent.pop(key)
             parent[key] = [item]
 
+    return target
+
+
+def dict_jsonify(source: dict,
+                 jsonify_keys: bool = False,
+                 jsonify_values: bool = True) -> dict:
+    """
+    Convert the *[key, value]* pairs in *source* into values that can be serialized to JSON, thus avoiding *TypeError*.
+
+    The parameters *jsonify_keys* and *jsonify_values* specify whether keys and values are to be *jsonified*,
+    respectively. It is expected that either *jsonify_keys* or *jsonify_values*, or both, is set to *True*.
+
+    Possible transformations of keys and values:
+      - *IntEnum* is changed with *int()*
+      - *bytes*, *bytearray*, and generic *Enum* are changed with *str()*
+      - *date* and *datetime* are changed to their *ISO* representations
+      - *Path* is changed to its *POSIX* representation
+      - *dict* is recursively *jsonified* with *dict_jsonify()* (values, only)
+      - *list* is recursively *jsonified* with *list_jsonify()* (values, only)
+      - all other types are left unchanged
+
+    Note that retrieving the original values through a reversal of this process is not deterministic.
+    The transformation is recursively carried out, that is, any *dict* or *list* set as value will be
+    *jsonified* accordingly. For convenience, the possibly modified *source* itself is returned.
+
+    HAZARD: depending on the type of object contained in *source*, the final result may not be serializable.
+
+    :param source: the dict to be made serializable
+    :param jsonify_keys: whether the keys in *source* should be *jsonified* (defaults to *False*)
+    :param jsonify_values: whether the values in *source* should be *jsonified* (defaults to *True*)
+    :return: the modified input *dict*
+    """
+    # traverse the input 'dict'
+    keys: list[Any] = []
+    for key, value in source.items():
+
+        # values transformations
+        if jsonify_values:
+            if isinstance(value, dict):
+                dict_jsonify(source=value,
+                             jsonify_keys=jsonify_keys,
+                             jsonify_values=jsonify_values)
+            elif isinstance(value, list):
+                from .list_pomes import list_jsonify
+                source[key] = list_jsonify(source=value)
+            elif isinstance(value, IntEnum):
+                source[key] = int(value)
+            elif isinstance(value, bytes | bytearray | Enum):
+                source[key] = str(value)
+            elif isinstance(value, date):
+                source[key] = value.isoformat()
+            elif isinstance(value, Path):
+                source[key] = value.as_posix()
+
+        # key transformations
+        if jsonify_keys and \
+                isinstance(key, Enum | Path | bytes | bytearray | date):
+            keys.append(key)
+
+    # transform the keys
+    for key in keys:
+        if isinstance(key, IntEnum):
+            source[int(key)] = source.pop(key)
+        elif isinstance(key, bytes | bytearray | Enum):
+            source[str(key)] = source.pop(key)
+        elif isinstance(key, date):
+            source[key.isoformat()] = source.pop(key)
+        elif isinstance(key, Path):
+            source[key.as_posix()] = source.pop(key)
+
+    return source
+
+
+def dict_hexify(source: dict,
+                hexify_keys: bool = False,
+                hexify_values: bool = True) -> dict:
+    """
+    Convert the *[key, value]* pairs in *source* to their hexadecimal representantions.
+
+    The parameters *hexify_keys* and *hexify_values* specify whether keys and values are to be *hexified*,
+    respectively. It is expected that either *hexify_keys* or *hexify_values*, or both, is set to *True*.
+
+    Possible transformations of keys and values:
+      - *str* is changed with *<value>.encode().hex()*
+      - *int* is changed with *float(<value>).hex()*
+      - *float*, *bytes*, and *bytearray* are changed using their built-in *hex()* method
+      - *date* and *datetime* are changed using their ISO representations
+      - *Path* is changed using its POSIX representation
+      - *dict* is recursively *hexified* with *dict_hexify()*
+      - *list* is recursively *hexified* with *list_hexify()*
+      - all other types are left unchanged
+
+    Note that retrieving the original values through a reversal of this process is not deterministic.
+    The transformation is recursively carried out, that is, any *dict* or *list* set as value will be
+    *hexified* accordingly. For convenience, the possibly modified *source* itself is returned.
+
+    :param source: the dict to be made serializable
+    :param hexify_keys: whether the keys in *source* should be *hexified* (defaults to *False*)
+    :param hexify_values: whether the values in *source* should be *hexified* (defaults to *True*)
+    :return: the modified input *dict*
+    """
+    # needed imports
+    from .list_pomes import list_hexify
+
+    # traverse the input 'dict'
+    keys: list[Any] = []
+    for key, value in source.items():
+
+        # values transformations
+        if hexify_values:
+            if isinstance(value, dict):
+                dict_hexify(source=value,
+                            hexify_keys=hexify_keys,
+                            hexify_values=hexify_values)
+            elif isinstance(value, list):
+                source[key] = list_hexify(source=value)
+            elif isinstance(value, str):
+                source[key] = value.encode().hex()
+            elif isinstance(value, int):
+                source[key] = float(value).hex()
+            elif isinstance(value, float | bytes | bytearray):
+                source[key] = value.hex()
+            elif isinstance(value, Path):
+                source[key] = value.as_posix().encode().hex()
+            elif isinstance(value, date):
+                source[key] = value.isoformat().encode().hex()
+
+        # key transformations
+        if hexify_keys and \
+                isinstance(key, dict | list | int | float | str | Path | bytes | bytearray | date):
+            keys.append(key)
+
+    # transform the keys
+    for key in keys:
+        if isinstance(key, str):
+            source[key.encode().hex()] = source.pop(key)
+        elif isinstance(key, int):
+            source[float(key).hex()] = source.pop(key)
+        elif isinstance(key, float | bytes | bytearray):
+            source[key.hex()] = source.pop(key)
+        elif isinstance(key, Path):
+            source[key.as_posix().encode().hex()] = source.pop(key)
+        elif isinstance(key, date):
+            source[key.isoformat().encode().hex()] = source.pop(key)
+
+    return source
+
 
 if __name__ == "__main__":
 
